/*
 * lib/inet6.c	This file contains an implementation of the "INET6"
 *		support functions for the net-tools.
 *		(NET-3 base distribution).
 *
 * Version:	lib/inet6.c 1.32 1996-08-16
 *
 * Author:	Fred N. van Kempen, <waltje@uwalt.nl.mugnet.org>
 *		Copyright 1993 MicroWalt Corporation
 *
 * Modified:
 *960113 {1.21}	Bernd Eckenfels :	rresolve cache bug.
 *960128 {1.22}	Bernd Eckenfels :	endian bug in print
 *960203 {1.23} Bernd Eckenfels :       net-features support
 *960217 {1.24} Bernd Eckenfels :	get_sname
 *960219 {1.25} Bernd Eckenfels :	extern int h_errno
 *960329 {1.26} Bernd Eckenfels :	resolve 255.255.255.255	
 *
 *		This program is free software; you can redistribute it
 *		and/or  modify it under  the terms of  the GNU General
 *		Public  License as  published  by  the  Free  Software
 *		Foundation;  either  version 2 of the License, or  (at
 *		your option) any later version.
 */
#include "config.h"

#if HAVE_AFINET6
#include <sys/types.h>
#include <linux/in.h>
#include <linux/socket.h>
#include <arpa/inet.h>
#include <arpa/nameser.h>
#include <linux/route.h>
#include <ctype.h>
#include <errno.h>
/*#include <netdb.h> FIXME*/
#include <netdb.h>
#include <resolv.h>
#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <unistd.h>
#include "version.h"
#include "net-support.h"
#include "pathnames.h"
#define  EXTERN
#include "net-locale.h"

extern int h_errno;  /* some netdb.h versions don't export this */

struct addr {
  struct sockaddr_in6	addr6;
  char			*name;
  struct addr		*next;
};

static struct addr *INET6_nn = NULL;	/* addr-to-name cache		*/

static int
INET6_resolve(char *name, unsigned int ipv4addr, struct sockaddr_in6 *sin6)
{
	struct hostent *host;
	unsigned int i;
	char*	semicolon;
	
	sin6->sin6_family = AF_INET6;
	sin6->sin6_port = 0;
	sin6->sin6_flowinfo = 0;
	
	/* It might be a numbered address */
	if (inet_pton(AF_INET6, name, &sin6->sin6_addr)==1)
		return(0);

	/* Default is special, meaning ::0 */
	if (!strcmp(name, "default")) {
		CLR_ADDR6(sin6->sin6_addr);
		return(1);
	}
	/* is it a IPv4 compatible address to find */
	
	if (semicolon = strstr (name, "::")) {
		name += 2; /* skip the semicolon */
		ipv4addr |= FLAG_IP4COMP;
	}

        if (host = gethostbyname2 (name, AF_INET6)) {
	   for (i = 0; host->h_addr_list[i]; i++) {
		if ((ipv4addr && IS_IPV4ADDR6( *(struct in6_addr*) host->h_addr_list[i])) ||
		    (!ipv4addr && !IS_IPV4ADDR6(*(struct in6_addr*) host->h_addr_list[i])))  { 
		   COPY_ADDR6(sin6->sin6_addr, *(struct in6_addr *) host->h_addr_list[i]);
		   return (0);
		} 
	   }
	   /* no ipv6 address found */
	   errno = NO_DATA;
	} else {
	    if (ipv4addr && (host = gethostbyname2 (name, AF_INET))) {
	    	CLR_ADDR6(sin6->sin6_addr); 
	    	sin6->sin6_addr.s6_addr32[3] = *(__u32 *) host->h_addr_list[0];
	    	return (0);
	    }
	    errno = h_errno;
	}
	return (-1);
}

static int
INET6_rresolve(char *name, struct sockaddr_in6 *sin6, int numeric)
{
	struct hostent *ent;
	struct addr *pn;
	struct in6_addr *ad;
	
	if (sin6->sin6_family != AF_INET6) {
#ifdef DEBUG
	fprintf(stderr, NLS_CATGETS(catfd, inetSet, inet_debug1, "rresolve: unsupport address family %d !\n"), sin6->sin6_family);
#endif
		errno = EAFNOSUPPORT;
		return(-1);
	}
	
	ad = &sin6->sin6_addr;
	ent = NULL;	
	
	if ((numeric & 0x7FFF) == 0) { /* no bit set = address name */
		if (numeric & FLAG_TARGET) { /* Destination target */
		        if (sin6->sin6_port == 0) { 
		          /* sin6_port is the address prefix length in this 
		             case only */
				strcpy(name, "default");
				return(0);
		        }     
	         } else if (IS_ANYADDR6(*ad)) { /* Gateway */
			        strcpy(name, "*");
				return(0);
		        }
		if (((numeric & FLAG_TARGET) && (sin6->sin6_port == 128)) ||
		     !(numeric & FLAG_TARGET)) {
		 /* search made for a host or a gateway else it is a network */
		    pn = INET6_nn;
		    while (pn != NULL) {
			if (SAME_ADDR6(pn->addr6.sin6_addr, *ad)) {
				strcpy(name, pn->name);
				return(0);
			}
			pn = pn->next;
		    }
	
		    ent = gethostbyaddr((char *) ad, IN6ADDRSZ, AF_INET6);
		    if (ent != NULL)
			 sprintf (name, "%s%s", IS_IPV4ADDR6(*ad)?"::":"", ent->h_name);
		}
	}
	
	if (ent == NULL)
		inet_ntop(AF_INET6, ad, name, INET6_ADDRSTRLEN);
	
	pn = (struct addr *)malloc(sizeof(struct addr));
	pn->addr6 = *sin6;
	pn->next = INET6_nn;
	pn->name = (char *) malloc(strlen(name) + 1);
	strcpy(pn->name, name);
	INET6_nn = pn;
	
	return(0);

}


static void
INET6_reserror(char *text)
{
  herror(text);
}

/* Display an Internet socket address. */
static char *
INET6_print(unsigned char *ptr)
{
  static char buff[128];

  return(inet_ntop(AF_INET6, 
                   ((struct in6_addr *) ptr), 
                   buff, INET6_ADDRSTRLEN));
}


/* Display a host name from an Internet socket address. */
static char *
INET6_sprint(struct sockaddr *sap, int numeric)
{
  static char buff[128];

  if (sap->sa_family == 0xFFFF || sap->sa_family == AF_UNSPEC)
    return(NLS_CATBUFF (catfd, inetSet, inet_none, "[NONE SET]", buff, 128));
  
  if (INET6_rresolve(buff, (struct sockaddr_in6 *) sap,  numeric) != 0)
	  return(NULL);
  return(buff);
}


/* Read Internet socket address from /proc/net6 */
static int
INET6_getsock(char *bufp, struct sockaddr *sap)
{
	struct sockaddr_in6 *sin6 = (struct sockaddr_in6 *) sap;

	if (sscanf(bufp, "%lx:%lx:%lx:%lx", &sin6->sin6_addr.s6_addr32[0],
		   &sin6->sin6_addr.s6_addr32[1],
		   &sin6->sin6_addr.s6_addr32[2],
		   &sin6->sin6_addr.s6_addr32[3]) != 4)
		return(-1);

	sin6->sin6_family = AF_INET6;
	sin6->sin6_flowinfo = 0;

	/* sin6_port contains prefix length 
	sin6->sin6_port = 0; */

	return(0); 
}

static int
INET6_input(int type, char *bufp, struct sockaddr *sap)
{
	switch(type) {
  	case FROM_FIB:
		return(INET6_getsock(bufp,sap));
	case FROM_DNS_4:
		return(INET6_resolve(bufp, FLAG_IP4COMP, (struct sockaddr_in6 *) sap));
	case FROM_DNS:
	default:
		return(INET6_resolve(bufp, 0, (struct sockaddr_in6 *) sap));
	}
}

struct aftype inet6_aftype = {
	"inet6",
	NULL, 
	AF_INET6,
	sizeof(struct sockaddr_in6),
        INET6_print,
	INET6_sprint,
	INET6_input,
	INET6_reserror,
	NULL			/*INET_rprint*/,
	NULL			/*INET_rinput*/
};


#endif	/* HAVE_AFINET6 */
